#include <cstdlib>
#include <time.h>
#include <math.h>

#include <iostream>
#include <algorithm>

#include <boost/program_options.hpp>

#include "MCL/core.h"
#include "MCL/scheduling/ws/ws.h"
#include "MCL/scheduling/pdf/pdf.h"
#include "MCL/array.h"
#include "Benchmark/merge_sort/STDMergeSort.h"

inline bool Cmp(size_t f, size_t s){
  return f<s;
}

inline bool LogCmp(size_t f, size_t s){
  return log(f)<log(s);
}

int main(int argc, char* argv[]) {
  namespace po = boost::program_options;
  // Declare the supported options.
  po::options_description desc("Allowed options");
  desc.add_options()
    ("help", "Produce help message")
    ("test", "Test that the result is sorted")
    ("threads", po::value<size_t>(), "Number of threads used")
    ("runsize", po::value<size_t>(), "Number of elements in each runs")
    ("runs", po::value<size_t>(), "Number of runs")
    ("tasks", po::value<size_t>(), "Number of tasks used for merging")
    ("PDF", "PDF scheduler is used")
    ("WS", "WS scheduler is used")
    ("log", "Logarithm comparison is used")
    ;
  po::variables_map args;
  po::store(po::parse_command_line(argc, argv, desc), args);
  po::notify(args);  
  if (args.count("help")) {
    std::cout << desc << "\n";
    return 1;
  }
  bool test = false;
  if(args.count("test"))
    test = true;
  size_t numThreads = 1;
  if(args.count("threads"))
    numThreads = args["threads"].as<size_t>();
  size_t runSize = 1024;
  if(args.count("runsize"))
    runSize = args["runsize"].as<size_t>();
  size_t numRuns = 32;
  if(args.count("runs"))
    numRuns = args["runs"].as<size_t>();
  size_t numTasks = 1;
  if(args.count("tasks"))
    numTasks = args["tasks"].as<size_t>();
  bool pdf = false;
  if(args.count("PDF"))
    pdf = true;
  bool ws = false;
  if(args.count("WS"))
    ws = true;
  bool log = false;
  if(args.count("log"))
    log = true;

  if(!ws && !pdf) {
    std::cout<<"No scheduler!"<<std::endl;
    return -1;
  } else if(ws && pdf){
    std::cout<<"Both scheduler slected"<<std::endl;
    return -1;
  }

  mcl::BasicArray<size_t> input(runSize*numRuns);
  mcl::BasicArray<size_t> output(runSize*numRuns);
  for(size_t i=0;i<input.size();++i){
    input[i] = i;
  }
  srand ( time(NULL) );
  std::random_shuffle(input.begin(),input.end());
  double mergeTime;
  double sortTime;

  if(ws){
    typedef mcl::LockingWorkStealingScheduler::Task Task;
    mcl::LockingWorkStealingScheduler* s = new mcl::LockingWorkStealingScheduler();
    mcl::ThreadPool<mcl::LockingWorkStealingScheduler>* tp = new mcl::ThreadPool<mcl::LockingWorkStealingScheduler>(*s,numThreads); 
    Task* top;
    if(log) {
      top = new STDMergeSort<mcl::BasicArray<size_t>,
			     std::function<bool(size_t,size_t)>,
			     Task>(&input,
				   &output,
				   runSize,
				   numTasks,
				   &mergeTime,
				   &sortTime,
				   LogCmp);
    } else {
      top = new STDMergeSort<mcl::BasicArray<size_t>,
			     std::function<bool(size_t,size_t)>,
			     Task>(&input,
				   &output,
				   runSize,
				   numTasks,
				   &mergeTime,
				   &sortTime,
				   Cmp);
  }
    tp->AddTask(top);
    tp->Start();
    delete top;
  } else {
    typedef mcl::LockingPDFScheduler::Task Task;
    mcl::LockingPDFScheduler* s = new mcl::LockingPDFScheduler();
    mcl::ThreadPool<mcl::LockingPDFScheduler>* tp = new  mcl::ThreadPool<mcl::LockingPDFScheduler>(*s,numThreads);
    Task* top;
    if(log) {
      top = new STDMergeSort<mcl::BasicArray<size_t>,
				   std::function<bool(size_t,size_t)>,
				   Task >(&input,
					  &output,
					  runSize,
					  numTasks,
					  &mergeTime,
					  &sortTime,
					  LogCmp);
    } else {
      top = new STDMergeSort<mcl::BasicArray<size_t>,
				   std::function<bool(size_t,size_t)>,
				   Task >(&input,
					  &output,
					  runSize,
					  numTasks,
					  &mergeTime,
					  &sortTime,
					  Cmp);
    }
    tp->AddTask(top);
    tp->Start();
    delete top;
  }
  std::cout<<mergeTime<<","<<sortTime;
  if(test) {
    for(size_t i=0;i<output.size();++i){
      if(output[i] != i){
	std::cout<<" Failed"<<std::endl;
	break;
      }
    }
    std::cout<<" Passed"<<std::endl;
  }
  return 0;
}
