#! /bin/bash
####################################################################
# benchmarks Mergsort speedup
# arguments are:
#  -o output filename
#  -w ws scheduler
#  -p pdf scheduler
#  -s runsize
#  -r number of runs
#  -t maximun number of threads 1,2,4...max is used
#  -a number of repetitions to average over
#  -l wether logarithm comparison is used 
###################################################################
MergeSort="./Mergesort"
seqSort="../sequential_sort"
logname="mergesort_speedup_data.csv"

ws=false
pdf=false
runsize=65536
runs=16
maxThreads=4
reps=3
tasks=4
log=""
while getopts "wpls:r:t:a:o:" opt; do 
    case $opt in
	w)
	    ws=true
	    ;;
	p)
	    pdf=true
       	    ;;
	s)
	    runsize=$OPTARG
	    ;;
	r)
	    runs=$OPTARG
	    ;;
	t)
	    maxThreads=$OPTARG
	    ;;
	a)
	    reps=$OPTARG
	    ;;
	o) 
	    logname=$OPTARG
	    ;;
	l)
	    log="--log"
	    ;;
    esac
done 

#clear the file
touch $logname
cat /dev/null > $logname

if $pdf
then
    heading="#scheduler, threads, runsize, numruns, tasks, repetitions, reps*(sorttime, mergetime)"
    echo $heading >> $logname
    thread=1
    until [ $thread -gt $maxThreads ]
    do
	rep=0
	line="PDF,${thread},${runsize},${runs},${thread},${reps}"
	while [ $rep -lt $reps ]
	do 
	    out="$( $MergeSort --PDF --threads $thread --runsize $runsize --runs $runs --tasks $thread $log )"
	    echo " $MergeSort --PDF --threads $thread --runsize $runsize --runs $runs --tasks $thread $log (run sort, merge) = $out " >&2
	    line="${line},$out"
	    rep=$(($rep +1))
	done
	echo $line >> $logname
	let thread=$thread+$thread
    done
fi
if $ws
then
    thread=1
    until [ $thread -gt $maxThreads ]
    do
	rep=0
	line="WS,${thread},${runsize},${runs},${thread},${reps}"
	while [ $rep -lt $reps ]
	do 
	    out="$( $MergeSort --WS --threads $thread --runsize $runsize --runs $runs --tasks $thread $log )"
	    echo " $MergeSort --WS --threads $thread --runsize $runsize --runs $runs --tasks $thread $log (run sort, merge) = $out " >&2
	    line="${line},$out"
	    rep=$(($rep +1))
	done
	echo $line >> $logname
	let thread=$thread+$thread
    done
fi
#run sequnetial
rep=0
inputsize=$(($runs * $runsize))
line="SEQ,1,${runsize},${runs},1,${reps}"
while [ $rep -lt $reps ]
do
    out="$($seqSort --input $inputsize $log)"
    line="${line},$out"
    echo "$seqSort --input $inputsize $log = $out" >&2
    rep=$(($rep +1))
done
echo $line >> $logname