#ifndef STD_MCL_MERGESORT_BENCHMARK
#define STD_MCL_MERGESORT_BENCHMARK

#include <iostream>

#include "MCL/array.h"
#include "MCL/algorithms/merge.h"
#include "MCL/algorithms/viewDivision.h"
#include "MCL/algorithms/sequentialSortTask.h"

template<class View, class Compare, class Task>
class STDMergeSort : public Task {
  typedef typename mcl::BasicArray<mcl::SubView<View> > TopView;
  typedef typename mcl::MultiWayMergeTask<TopView, View, Compare,Task> MergeTask;
  typedef typename mcl::SortTask<View,Compare,Task> RunSortTask;
public:
  STDMergeSort(View* i, 
	       View* o,
	       size_t runSize,
	       size_t numTasks,
	       double* mergeTime,
	       double* sortTime,
	       Compare cmp)
    : input_(i), output_(o), runSize_(runSize), numTasks_(numTasks), mergeTime_(mergeTime), sortTime_(sortTime), cmp_(cmp) { }

  void operator()(){
    struct timespec start, end;
    TopView Runs = mcl::BlocksOfAtMost(*input_,runSize_);
    mcl::BasicArray<Task*> sortTasks(Runs.size());
    for(size_t i=0;i<Runs.size();++i){
      sortTasks[i] = new RunSortTask(Runs[i],cmp_);
    }
    clock_gettime(CLOCK_MONOTONIC, &start);
    Spawn(sortTasks);
    clock_gettime(CLOCK_MONOTONIC, &end);
    *mergeTime_ = (end.tv_sec - start.tv_sec);
    *mergeTime_ += (end.tv_nsec - start.tv_nsec) / 1000000000.0;

    for(size_t i=0;i<Runs.size();++i) {
      delete sortTasks[i];
    }
    Task* task = new MergeTask(&Runs,output_,numTasks_,cmp_);
    clock_gettime(CLOCK_MONOTONIC, &start);
    Spawn(task);
    clock_gettime(CLOCK_MONOTONIC, &end);
    *sortTime_ = (end.tv_sec - start.tv_sec);
    *sortTime_ += (end.tv_nsec - start.tv_nsec) / 1000000000.0;
    delete task;
  }

protected:
  View* input_;
  View* output_;
  size_t runSize_;
  size_t numTasks_;
  double* mergeTime_;
  double* sortTime_;
  Compare cmp_;
};

#endif
