#include <stdlib.h>
#include <time.h>

#include <iostream>
#include <algorithm>
#include <functional>
#include <string>

#include <boost/program_options.hpp>

#include "MCL/core.h"
#include "MCL/scheduling/ws/ws.h"
#include "MCL/scheduling/pdf/pdf.h"
#include "MCL/array.h"
#include "MCL/algorithms/QuickSort.h"

inline bool Cmp(size_t f, size_t s){
  return f<s;
}

inline bool LogCmp(size_t f, size_t s){
  return log(f)<log(s);
}

int main(int argc, char* argv[]) {
  //argument parsing
  namespace po = boost::program_options;
  po::options_description desc("Allowed options");
  desc.add_options()
    ("help", "produce help message")
    ("test", "Test that the result is sorted")
    ("grain", po::value<size_t>(), "Number of elements sorted directly")
    ("blocksize", po::value<size_t>(), "Number of elements in each block")
    ("tasks", po::value<size_t>(), "Maximal number of tasks used for parallel pratitioning")
    ("threads", po::value<size_t>(), "Number of threads used")
    ("input", po::value<size_t>(), "Number elements sorted")
    ("WS", "WS scheduler used")
    ("PDF", "PDF scheduler used")
    ("log", "Indicate that logarithm comparison shold be used")
    ;
  po::variables_map args;
  po::store(po::parse_command_line(argc, argv, desc), args);
  po::notify(args);  
  if (args.count("help")) {
    std::cout << desc << "\n";
    return 1;
  }
  bool test = false;
  if(args.count("test"))
    test = true;
  size_t grainSize = 1024;
  if(args.count("grain"))
    grainSize = args["grain"].as<size_t>();
  size_t blockSize = 512;
  if(args.count("blocksize"))
    blockSize = args["blocksize"].as<size_t>();
  size_t numTasks = 2;
  if(args.count("tasks"))
    numTasks = args["tasks"].as<size_t>();
  size_t numThreads = 2;
  if(args.count("threads"))
    numThreads = args["threads"].as<size_t>();
  size_t inputSize = 65536;
  if(args.count("input"))
    inputSize = args["input"].as<size_t>();
  bool pdf = false;
  if(args.count("PDF"))
    pdf = true;
  bool ws = false;
  if(args.count("WS"))
    ws = true;
  bool log = false;
  if(args.count("log"))
     log = true;

  if(!ws && !pdf) {
    std::cout<<"No scheduler!"<<std::endl;
    return -1;
  } else if(ws && pdf){
    std::cout<<"Both scheduler slected"<<std::endl;
    return -1;
  }

  //genrate input
  mcl::BasicArray<size_t> data(inputSize);
  for(size_t i=0;i<inputSize;++i){
    data[i] = i;
  }
  srand ( time(NULL) );
  std::random_shuffle(data.begin(),data.end());

  struct timespec start, end;
  //start scheduling
  if(ws) {
      typedef mcl::LockingWorkStealingScheduler::Task Task;
      mcl::LockingWorkStealingScheduler* s = new mcl::LockingWorkStealingScheduler();
      mcl::ThreadPool<mcl::LockingWorkStealingScheduler>* tp = new mcl::ThreadPool<mcl::LockingWorkStealingScheduler>(*s,numThreads);
      Task* top;
      if(log) {
	top = new mcl::QuickSortTask<mcl::BasicArray<size_t>,std::function<bool(size_t,size_t)>,Task>(&data,grainSize,blockSize,numTasks,LogCmp);
      } else {
	top = new mcl::QuickSortTask<mcl::BasicArray<size_t>,std::function<bool(size_t,size_t)>,Task>(&data,grainSize,blockSize,numTasks,Cmp);
      }
      tp->AddTask(top);
      clock_gettime(CLOCK_REALTIME, &start);
      tp->Start();
      clock_gettime(CLOCK_REALTIME, &end);
      delete top;
    } else {
      typedef mcl::LockingPDFScheduler::Task Task;
      mcl::LockingPDFScheduler* s = new mcl::LockingPDFScheduler();
      mcl::ThreadPool<mcl::LockingPDFScheduler>* tp = new  mcl::ThreadPool<mcl::LockingPDFScheduler>(*s,numThreads);
      Task* top;
      if(log) {
	top = new mcl::QuickSortTask<mcl::BasicArray<size_t>,std::function<bool(size_t,size_t)>,Task>(&data,grainSize,blockSize,numTasks,LogCmp);
      } else {
	top = new mcl::QuickSortTask<mcl::BasicArray<size_t>,std::function<bool(size_t,size_t)>,Task>(&data,grainSize,blockSize,numTasks,Cmp);
      }
      tp->AddTask(top);
      clock_gettime(CLOCK_REALTIME, &start);
      tp->Start();
      clock_gettime(CLOCK_REALTIME, &end);
      delete top;
    } 

    double sortTime = (end.tv_sec - start.tv_sec);
    sortTime += (end.tv_nsec - start.tv_nsec) / 1000000000.0;
    std::cout<<sortTime;
    if(test) {
      for(size_t i=0;i<inputSize;++i){
	if(data[i] != i){
	  std::cout<<" Failed"<<std::endl;
	  break;
	}
      }
      std::cout<<" Passed"<<std::endl;
    }
    return 0;
    }
