#ifndef MCL_MUTEX
#define MCL_MUTEX

#include <pthread.h>
#include <tbb/atomic.h>
//#include <iostream>

class Mutex {
 public:
  pthread_mutex_t mutex_;
  pthread_cond_t condt_wait_;
  int numWaiting_;

  //requires mutex to be locked
  void Wait(){
    ++numWaiting_;
    pthread_cond_wait(&condt_wait_,&mutex_);
    --numWaiting_;
  }
  //requires mutex to be locked
  void Signal(){
    if(numWaiting_ > 0) {
      pthread_cond_signal(&condt_wait_);
    }
  }
  void Brodcast(){
    if(numWaiting_ > 0) {
      pthread_cond_broadcast(&condt_wait_);
    }
  }
  class Lock {
  public:
    Lock(Mutex* m) : mutex_(m) {
      pthread_mutex_lock(&mutex_->mutex_);
    }
    ~Lock(){
      pthread_mutex_unlock(&mutex_->mutex_);
    }
    void Wait(){
      mutex_->Wait();
    }
    void Signal(){
      mutex_->Signal();
    }
    void Brodcast(){
      mutex_->Brodcast();
    }
  private:
    Mutex* mutex_;
  };

  Lock lock(){
    return Lock(this);
  }

 Mutex() : numWaiting_(0) {
    pthread_mutex_init(&mutex_, NULL);
    pthread_cond_init (&condt_wait_, NULL);
  }
  ~Mutex() {
    pthread_mutex_destroy(&mutex_);
    pthread_cond_destroy(&condt_wait_);
  }
};

#endif //MCL_MUTEX
